
#include "Shot.h"
#include "Allocator.h"
#include "VisualizeParam.h"
#include "GameDefine.h"

#include "Juna/D2D/D2D_Blit.h"

#include <new>
#include <float.h>

Shot::Shot(Generator &generator,
		   const Point &position,float angle,const Vector & velocity)
  : I_Entity(generator.Entity),
	I_Visual(generator.Visual,VL_Top),
	I_Attacker(generator.Attacker),
	Position(position), Angle(angle), Velocity(velocity),
	Range(4), Power(100), LifeCounter(60)
{
}

Shot *Shot::Create(Generator &generator,const Point &position,float angle,const Vector & velocity)
{
	I_Allocator &a = generator.Allocator;
	assert(a.blockSize() >= sizeof(Shot) + sizeof(I_Allocator *));
	if (a.full())
		return 0;
	I_Allocator **hold = static_cast<I_Allocator **>(a.Allocate());
	*hold = &a;
	return new(hold + 1) Shot(generator,position,angle,velocity);
}
void Shot::Break(void)
{
	I_Allocator **hold = reinterpret_cast<I_Allocator **>(this) - 1;
	this->~Shot();
	(*hold)->Deallocate(hold);
}

bool Shot::Update(void)
{
	LifeCounter--;
	Position += Velocity;
	return LifeCounter > 0;
}
void Shot::Attack(const AttackParam &param)
{
	struct Circle : public Juna::D2::Intersection::I_Circle
	{
		Circle(const Shot &shot) : shot(shot) {}
		virtual Data GetCircleData(void) const
		{
			return Data(shot.position(),shot.range());
		}
		const Shot &shot;
	};
	Circle c(*this);

	typedef AttackParam::list list;
	const list &enemy = param.Enemy;
	for (list::const_iterator it = enemy.begin();it != enemy.end();++it)
	{
		if (it->attackedShape().Test(c))
		{
			Damage(*it,Power,Vector(0,0));
			LifeCounter = 0;
		}
	}
}
void Shot::Visualize(const VisualizeParam &param) const
{
	param.Screen.Blit(Position.x,Position.y,
		param.Sprite,Rect(128,0,Size(16,16)),Angle);
}



PiercingShot::PiercingShot(Generator &generator,
		   const Point &position,float angle,const Vector & velocity)
  : I_Entity(generator.Entity),
	I_Visual(generator.Visual,VL_Top),
	I_Attacker(generator.Attacker),
	Position(position), Angle(angle), Velocity(velocity),
	Range(4), Power(20), LifeCounter(60)
{
}

PiercingShot *PiercingShot::Create(Generator &generator,const Point &position,float angle,const Vector & velocity)
{
	I_Allocator &a = generator.Allocator;
	assert(a.blockSize() >= sizeof(PiercingShot) + sizeof(I_Allocator *));
	if (a.full())
		return 0;
	I_Allocator **hold = static_cast<I_Allocator **>(a.Allocate());
	*hold = &a;
	return new(hold + 1) PiercingShot(generator,position,angle,velocity);
}
void PiercingShot::Break(void)
{
	I_Allocator **hold = reinterpret_cast<I_Allocator **>(this) - 1;
	this->~PiercingShot();
	(*hold)->Deallocate(hold);
}

bool PiercingShot::Update(void)
{
	LifeCounter--;
	Position += Velocity;
	return LifeCounter > 0;
}
void PiercingShot::Attack(const AttackParam &param)
{
	struct Circle : public Juna::D2::Intersection::I_Circle
	{
		Circle(const PiercingShot &shot) : shot(shot) {}
		virtual Data GetCircleData(void) const
		{
			return Data(shot.position(),shot.range());
		}
		const PiercingShot &shot;
	};
	Circle c(*this);

	typedef AttackParam::list list;
	const list &enemy = param.Enemy;
	for (list::const_iterator it = enemy.begin();it != enemy.end();++it)
	{
		if (it->attackedShape().Test(c))
		{
			Damage(*it,Power,Vector(0,0));
		}
	}
}
void PiercingShot::Visualize(const VisualizeParam &param) const
{
	param.Screen.Blit(Position.x,Position.y,
		param.Sprite,Rect(128,0,Size(16,16)),Angle);
}




LaserShot::LaserShot(::Generator &generator,
		   const Point &position,float angle)
  : I_Entity(generator.Entity),
	I_Visual(generator.Visual,VL_Top),
	I_Attacker(generator.Attacker),
	Position(position), Direction(angle), Length(1024), Width(4), 
	Power(4), LifeCounter(60),
	Generator(generator)
{
}

LaserShot *LaserShot::Create(::Generator &generator,const Point &position,float angle)
{
	I_Allocator &a = generator.Allocator;
	assert(a.blockSize() >= sizeof(LaserShot));
	if (a.full())
		return 0;
	return new(a.Allocate()) LaserShot(generator,position,angle);
}
void LaserShot::Break(void)
{
	I_Allocator &a = Generator.Allocator;
	this->~LaserShot();
	a.Deallocate(this);
}

bool LaserShot::Update(void)
{
	return false;
}


namespace {
bool HitCheck(const Juna::D2::Intersection::I_Segment &line,const AttackParam::list &enemy)
{
	for (AttackParam::list::const_iterator it = enemy.begin();it != enemy.end();++it)
		if (it->attackedShape().Test(line))
			return true;
	return false;
}
}//	anonymous namespace

void LaserShot::Attack(const AttackParam &param)
{
	struct Line : public Juna::D2::Intersection::I_Segment
	{
		virtual Data GetSegmentData(void) const
		{
			return Data(p1,p2,half_width);
		}
		Point p1,p2;
		float half_width;
	};
	Line l;
	Vector v(Direction.cos() * Length,Direction.sin() * Length);
	l.p1 = Position;
	l.p2 = Position + v;
	l.half_width = Width;

	const AttackParam::list &enemy = param.Enemy;

	if (HitCheck(l,enemy))
	{
		float length = Length / 2;
		v /= 2;
		l.p2 = l.p1 + v;
		for (int i = 0;i < 10;i++)
		{
			if (HitCheck(l,enemy))
			{
				Length -= length;
			}
			else
			{
				l.p1 = l.p2;
			}
			v /= 2;
			l.p2 = l.p1 + v;
			length /= 2;
		}
	}

	v.Set(Direction.cos() * Length,Direction.sin() * Length);
	l.p1 = Position;
	l.p2 = Position + v;
	for (AttackParam::list::const_iterator it = enemy.begin();it != enemy.end();++it)
	{
		if (it->attackedShape().Test(l))
		{
			Damage(*it,Power,Vector(0,0));
		}
	}
}



void LaserShot::Visualize(const VisualizeParam &param) const
{
	using namespace Juna::D2D;
	I_ScreenSurface &screen = param.Screen;

	screen.SetBlendType(BT_ADD);
	LinePoint lp[2];
	lp[0].Set(Position.x,Position.y,16,ARGB(255,255,255,255));
	lp[1].Set(Position.x + Direction.cos() * Length,
			  Position.y + Direction.sin() * Length,16,ARGB(255,255,255,255));
	Blitter::LineBlit(screen,lp,2,param.Sprite,FRect(176,0,FSize(16,16)));
	screen.SetBlendType(BT_ALPHA);
}


